# -*- coding: utf-8 -*-
import logging
from requests.exceptions import HTTPError

from django.utils.translation import ugettext_lazy as _
from rest_framework.authentication import TokenAuthentication
from rest_framework.exceptions import AuthenticationFailed

from .token_info import get_token_info
from .settings import am_auth_settings


logger = logging.getLogger(__name__)


class AMOAuth2TokenAuthentication(TokenAuthentication):
    @property
    def keyword(self):
        return am_auth_settings.TOKEN_TYPE

    def authenticate_credentials(self, key):
        try:
            logger.debug('Trying to get token info for %s', key)
            token_info = get_token_info(key)
        except HTTPError as e:
            logger.debug(e)
            if e.response.status_code == 400:
                raise AuthenticationFailed(_('Invalid token.'))
            if e.response.status_code in (401, 403):
                raise AuthenticationFailed(_(
                    'Cannot authorize the service on AccountManager.'
                ))
            if 500 <= e.response.status_code <= 599:
                raise AuthenticationFailed(_(
                    'Authorization server responded with HTTP 5xx.'
                ))
            else:
                raise e
        return (token_info.user, token_info)



# -*- coding: utf-8 -*-
import json
import logging
from urlparse import urljoin

from django.core.cache import cache
from django.contrib.auth.models import Permission
from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session
from requests.auth import HTTPBasicAuth

from .settings import am_auth_settings


logger = logging.getLogger(__name__)


class AccountManagerClient(object):
    def __init__(self):
        token = cache.get(am_auth_settings.TOKEN_CACHE_KEY)
        self.client = OAuth2Session(
            client=BackendApplicationClient(
                client_id=self.client_id,
            ),
            token=token,
        )
        if not token:
            logger.debug(
                "Token has not found in the cache (key='%s'). "
                "Requesting it from AccountManager.",
                am_auth_settings.TOKEN_CACHE_KEY
            )
            self.fetch_token()

    def fetch_token(self):
        token = self.client.fetch_token(
            token_url=self._token_url,
            auth=HTTPBasicAuth(self.client_id, self.client_secret),
            verify=am_auth_settings.VERIFY_SSL_CERTIFICATE,
        )
        cache.set(am_auth_settings.TOKEN_CACHE_KEY, token, token['expires_in'])

    def __del__(self):
        self.client.close()

    @property
    def client_id(self):
        return am_auth_settings.OAUTH2_CLIENT_ID

    @property
    def client_secret(self):
        return am_auth_settings.OAUTH2_CLIENT_SECRET

    @property
    def api_url(self):
        return am_auth_settings.API_URL

    @property
    def _token_url(self):
        return urljoin(self.api_url, '/oauth2/v1/token/')

    @property
    def _token_introspection_url(self):
        return urljoin(self.api_url, '/oauth2/v1/introspect-token/')

    def request(self, method, url, **kwargs):
        request_method = getattr(self.client, method)
        response = request_method(url, **kwargs)
        if response.status_code in (401, 403):
            logger.info(
                'Response code is %s (%s). '
                'Trying to obtain a new token and make request again.',
                response.status_code,
                response.content,
            )
            self.fetch_token()
            response = request_method(url, **kwargs)
        response.raise_for_status()
        return response.json()

    def introspect_token(self, token):
        return self.request(
            'post',
            self._token_introspection_url,
            data=json.dumps({'token': token}),
            verify=am_auth_settings.VERIFY_SSL_CERTIFICATE,
            headers={
                'content-type': 'application/json',
                'accept': 'application/json',
            },
        )

    @property
    def _synchronize_permissions_url(self):
        return urljoin(
            self.api_url,
            '/service-api/v1/services/{service_id}'
            '/synchronize-permissions/'.format(
                service_id=am_auth_settings.SERVICE_ID,
            )
        )

    def synchronize_permissions(self):
        permissions = Permission.objects.all()
        request_data = {
            'data': {
                'attributes': {
                    'permissions': [
                        {
                            'codename': '{app_label}.{codename}'.format(
                                app_label=permission.content_type.app_label,
                                codename=permission.codename
                            ),
                            'name': permission.name,
                        } for permission in permissions
                        if (
                            not am_auth_settings.SYNC_APPS or
                            permission.content_type.app_label
                           )
                        in am_auth_settings.SYNC_APPS
                    ],
                },
                'type': 'services',
            }
        }
        return self.request(
            'post',
            self._synchronize_permissions_url,
            data=json.dumps(request_data),
            headers={
                'content-type': 'application/vnd.api+json',
            },
            verify=am_auth_settings.VERIFY_SSL_CERTIFICATE,
        )



# -*- coding: utf-8 -*-
from rest_framework.permissions import DjangoModelPermissions, SAFE_METHODS


class AMPermissions(DjangoModelPermissions):
    def has_permission(self, request, view):
        if request.method in SAFE_METHODS:
            return True
        # Workaround to ensure DjangoModelPermissions are not applied
        # to the root view when using DefaultRouter.
        if getattr(view, '_ignore_model_permissions', False):
            return True

        if hasattr(view, 'get_queryset'):
            queryset = view.get_queryset()
        else:
            queryset = getattr(view, 'queryset', None)

        assert queryset is not None, (
            'Cannot apply DjangoModelPermissions on a view that '
            'does not set `.queryset` or have a `.get_queryset()` method.'
        )

        perms = self.get_required_permissions(request.method, queryset.model)

        return (
            request.user and
            (request.auth or not self.authenticated_users_only) and
            request.auth.user_has_perms(perms)
        )



# -*- coding: utf-8 -*-
import os
import urlparse

from django.conf import settings


DEFAULTS = {
    'API_URL': None,
    'OAUTH2_CLIENT_ID': None,
    'OAUTH2_CLIENT_SECRET': None,
    'SERVICE_ID': None,
    'VERIFY_SSL_CERTIFICATE': True,
    'TOKEN_INFO_TIMEOUT': 60,
    'SYNC_APPS': [],
    'TOKEN_CACHE_KEY': 'account-manager-service-token',
    'TOKEN_INFO_CACHE_KEY': 'account-manager-token-{token}',
    'TOKEN_TYPE': 'Bearer',
}

MANDATORY = (
    'API_URL',
    'OAUTH2_CLIENT_ID',
    'OAUTH2_CLIENT_SECRET',
    'SERVICE_ID',
    'TOKEN_TYPE',
    'SYNC_APPS',
)


class AMAuthSettings(object):
    def __init__(self, defaults=None, mandatory=None):
        self.user_settings = getattr(settings, 'REST_FRAMEWORK_AM_AUTH', None)
        self.defaults = defaults or {}
        self.mandatory = mandatory or ()

    def __getattr__(self, attr):
        if attr not in self.defaults.keys():
            raise AttributeError("Invalid AMAuth setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        self.validate_setting(attr, val)

        # Cache the result
        setattr(self, attr, val)
        if attr == 'API_URL':
            self.handle_api_url(val)
        return val

    def validate_setting(self, attr, val):
        if not val and attr in self.mandatory:
            raise AttributeError("AMAuth setting: '%s' is mandatory" % attr)

    def handle_api_url(self, api_url):
        parsed_url = urlparse.urlparse(api_url)
        if parsed_url.scheme == 'http':
            os.environ['OAUTHLIB_INSECURE_TRANSPORT'] = '1'


am_auth_settings = AMAuthSettings(DEFAULTS, MANDATORY)



# -*- coding: utf-8 -*-
import logging

from django.contrib.auth import get_user_model
from django.utils.functional import cached_property
from django.core.cache import cache

from .account_manager_client import AccountManagerClient
from .settings import am_auth_settings


logger = logging.getLogger(__name__)


def get_token_info(token):
    token_info_cache_key = am_auth_settings.TOKEN_INFO_CACHE_KEY.format(
        token=token,
    )
    if am_auth_settings.TOKEN_INFO_TIMEOUT:
        logger.debug(
            "Trying to get TokenInfo from the cache (key='%s').",
            token_info_cache_key,
        )
        token_info = cache.get(token_info_cache_key)
    else:
        logger.debug('Caching TokenInfo is disabled.')
        token_info = None
    if not token_info:
        logger.debug(
            'TokenInfo has not gotten from the cache. '
            'Requesting it from AccountManager.'
        )
        account_manager_client = AccountManagerClient()
        token_info = account_manager_client.introspect_token(token)
        if am_auth_settings.TOKEN_INFO_TIMEOUT:
            logger.debug(
                'Caching TokenInfo %s with timeout %s.',
                token_info_cache_key,
                am_auth_settings.TOKEN_INFO_TIMEOUT,
            )
            cache.set(
                token_info_cache_key,
                token_info,
                am_auth_settings.TOKEN_INFO_TIMEOUT,
            )
    return AMOAuth2TokenInfo(token, token_info)


class AMOAuth2TokenInfo(object):
    def __init__(self, token, token_info):
        self._token_info = token_info
        self.token = token

    @cached_property
    def user(self):
        UserModel = get_user_model()
        user, _created = UserModel.objects.get_or_create(
            username=self._token_info['username'],
        )
        return user

    def user_has_perms(self, perms):
        return all(
            perm in self._token_info['scope']
            for perm in perms
        )



from setuptools import setup

setup(
    name='djangorestframework-amauth',
    version='1.0',
    packages=['rest_framework_am_auth'],
    url='',
    license='',
    author='GS Labs',
    author_email='',
    description='AccountManager auth backend for DRF.',
    install_requires=(
        'Django',
        'djangorestframework',
        'requests_oauthlib',
    )
)
